import asyncio
from typing import Any
from typing_extensions import override

from aioimaplib import aioimaplib
from nonebot.drivers import Driver
from nonebot.utils import escape_tag
from nonebot.compat import model_dump

from nonebot.adapters import Adapter as BaseAdapter

from .bot import Bot
from .log import log
from .config import Config, BotInfo
from .event import NewMailMessageEvent
from .exception import ApiNotAvailable

CHECK_MAIL_INTERVAL = 3.0
RECONNECT_INTERVAL = 3.0


class Adapter(BaseAdapter):
    @override
    def __init__(self, driver: Driver, **kwargs: Any) -> None:
        super().__init__(driver, **kwargs)
        self.mail_config: Config = Config(**model_dump(self.config))
        self.tasks: set[asyncio.Task] = set()
        self.setup()

    @classmethod
    @override
    def get_name(cls) -> str:
        return "Mail"

    def setup(self) -> None:
        self.on_ready(self.startup)
        self.driver.on_shutdown(self.shutdown)

    async def startup(self) -> None:
        for bot in self.mail_config.mail_bots:
            task = asyncio.create_task(self.run_bot(bot))
            task.add_done_callback(self.tasks.discard)
            self.tasks.add(task)

    async def shutdown(self) -> None:
        for task in self.tasks:
            if not task.done():
                task.cancel()
        await asyncio.gather(
            *(asyncio.wait_for(task, timeout=10) for task in self.tasks),
            return_exceptions=True,
        )

    async def run_bot(self, bot_info: BotInfo) -> None:
        bot = Bot(self, bot_info.id, bot_info)
        task = asyncio.create_task(self._check_mailbox(bot))
        task.add_done_callback(self.tasks.discard)
        self.tasks.add(task)

    async def _check_mailbox(self, bot: Bot) -> None:
        bot_info = bot.bot_info
        imap_info = bot_info.imap
        if imap_info.tls:
            bot.imap_client = aioimaplib.IMAP4_SSL(
                host=imap_info.host, port=imap_info.port
            )
        else:
            bot.imap_client = aioimaplib.IMAP4(host=imap_info.host, port=imap_info.port)
        while True:
            try:
                # connect to the server
                if not await bot.login():
                    return
                self.bot_connect(bot)
                log(
                    "INFO",
                    f"<y>Bot {escape_tag(bot.self_id)}</y> connected",
                )
                # select the mailbox
                if not await bot.select_mailbox():
                    continue
                # infinite loop to check for new mails
                while True:
                    try:
                        await self._fetch_new_mail(bot)
                    except Exception as e:
                        log(
                            "ERROR",
                            (
                                f"<y>Bot {escape_tag(bot.self_id)}</y> "
                                "<r><bg #f8bbd0>"
                                "error while fetching mail"
                                "</bg #f8bbd0></r>"
                            ),
                            e,
                        )
                    await asyncio.sleep(CHECK_MAIL_INTERVAL)
            except Exception as e:
                log(
                    "ERROR",
                    (
                        f"<y>Bot {escape_tag(bot.self_id)}</y> "
                        "<r><bg #f8bbd0>"
                        "error while setting up, reconnecting..."
                        "</bg #f8bbd0></r>"
                    ),
                    e,
                )
            finally:
                # logout from the server
                await bot.logout()
                # disconnect the bot instance
                if bot.self_id in self.bots:
                    self.bot_disconnect(bot)
            await asyncio.sleep(RECONNECT_INTERVAL)

    async def _fetch_new_mail(self, bot: Bot):
        if bot.mailbox != "INBOX":
            log(
                "TRACE",
                f"<y>Bot {escape_tag(bot.self_id)}</y> is not in INBOX, skip fetching",
            )
            return
        if bot.readonly:
            log(
                "TRACE",
                f"<y>Bot {escape_tag(bot.self_id)}</y> is readonly, skip fetching",
            )
            return
        # Search for unseen mails
        mails = await bot.fetch_unseen_mail_list()
        for mail in mails:
            # Handle the mail as an event
            event = NewMailMessageEvent(**model_dump(mail))
            task = asyncio.create_task(bot.handle_event(event))
            task.add_done_callback(self.tasks.discard)
            self.tasks.add(task)

    @override
    async def _call_api(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, bot: Bot, api: str, **data: Any
    ) -> Any:
        log("DEBUG", (f"Bot {bot.bot_info.id} calling API <y>{api}</y>"))
        api_handler = getattr(bot.__class__, api, None)
        if api_handler is None:
            raise ApiNotAvailable
        return await api_handler(bot, **data)



import asyncio
import email.message
import email.mime.multipart
from typing_extensions import override
from typing import TYPE_CHECKING, Any, Union, NoReturn, Optional

import aioimaplib
import aiosmtplib
from nonebot.utils import escape_tag
from nonebot.message import handle_event

from nonebot.adapters import Bot as BaseBot

from .log import log
from .model import Mail
from .config import BotInfo
from .message import Message, MessageSegment
from .event import Event, MessageEvent, NewMailMessageEvent
from .utils import parse_byte_mail, escape_bytelines, extract_mail_parts
from .exception import ActionFailed, NetworkError, UninitializedException

if TYPE_CHECKING:
    from .adapter import Adapter


def _check_to_me(
    bot: "Bot",
    event: MessageEvent,
):
    if isinstance(event, NewMailMessageEvent) and bot.bot_info.id in {
        i.id for i in event.recipients_to
    }:
        event.to_me = True


async def _check_reply(
    bot: "Bot",
    event: MessageEvent,
):
    if not isinstance(event, NewMailMessageEvent) or event.in_reply_to is None:
        return
    try:
        event.reply = await bot.fetch_mail_of_id(
            mail_id=event.in_reply_to,
        )
        if event.reply and event.reply.sender.id == bot.bot_info.id:
            event.to_me = True
    except Exception as e:
        log(
            "WARNING",
            (
                f"<y>Bot {escape_tag(bot.self_id)}</y> "
                "failed to fetch the reply mail."
            ),
            e,
        )


class Bot(BaseBot):
    @override
    def __init__(self, adapter: "Adapter", self_id: str, bot_info: BotInfo):
        super().__init__(adapter, self_id)
        self.bot_info: BotInfo = bot_info
        self.imap_client: Optional[aioimaplib.IMAP4] = None
        self.mailbox: Optional[str] = None
        self.readonly: bool = False

    @override
    def __getattr__(self, name: str) -> NoReturn:
        raise AttributeError(
            f'"{self.__class__.__name__}" object has no attribute "{name}"'
        )

    @property
    def type(self) -> str:
        return "Mail"

    @override
    async def send(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        event: Event,
        message: Union[str, Message, MessageSegment],
        reply: Optional[bool] = None,
        **kwargs,
    ) -> Any:
        """
        Send a message to the event.

        - `event`: The event to reply.
        - `message`: The message to send.
        - `reply`: Whether using default reply settings.
        - `kwargs`: Other arguments.
            - `cc`: The list of CC recipients.
            - `bcc`: The list of BCC recipients.
            - `subject`: The subject of the mail.
            - `in_reply_to`: The Message ID of the mail to reply to.
            - `references`: The list of Message IDs of the mails in the thread.
            - `reply_to`: The list of addresses to recommend recipients to reply to.
        """
        if not isinstance(event, MessageEvent):
            raise RuntimeError("Event cannot be replied to!")
        if isinstance(message, str):
            message = Message([MessageSegment.text(message)])
        elif isinstance(message, MessageSegment):
            message = Message([message])
        if "reply" in message:
            reply_id = message["reply", 0].data.get("id")
            if reply_id == event.id:
                reply = True if reply is None else reply
                message = message.exclude("reply")
        if reply:
            if kwargs.get("subject") is None:
                kwargs["subject"] = f"Re: {event.subject}"
            if kwargs.get("in_reply_to") is None:
                kwargs["in_reply_to"] = event.id
            if kwargs.get("references") is None:
                kwargs["references"] = [event.id]
        await self.send_mail(
            message=message,
            recipient=event.sender.id,
            **kwargs,
        )

    async def send_to(
        self,
        recipient: Union[str, list[str]],
        message: Union[str, Message, MessageSegment],
        **kwargs,
    ) -> None:
        """
        Send a mail to the given recipient(s).

        - `recipient`: The recipient(s) to send the mail to.
        - `message`: The message to send.
        - `kwargs`: Other arguments.
            - `cc`: The list of CC recipients.
            - `bcc`: The list of BCC recipients.
            - `subject`: The subject of the mail.
            - `in_reply_to`: The Message ID of the mail to reply to.
            - `references`: The list of Message IDs of the mails in the thread.
            - `reply_to`: The list of addresses to recommend recipients to reply to.
        """
        recipient = [recipient] if isinstance(recipient, str) else recipient
        if isinstance(message, str):
            message = Message([MessageSegment.text(message)])
        elif isinstance(message, MessageSegment):
            message = Message([message])
        await self.send_mail(
            message=message,
            recipient=recipient,
            **kwargs,
        )

    async def send_mail(
        self,
        message: Union[Message, email.message.EmailMessage],
        recipient: Union[str, list[str], None] = None,
        cc: Union[str, list[str], None] = None,
        bcc: Union[str, list[str], None] = None,
        subject: Optional[str] = None,
        in_reply_to: Optional[str] = None,
        references: Union[str, list[str], None] = None,
        reply_to: Union[str, list[str], None] = None,
    ) -> None:
        """
        Send a mail to the given recipients.

        - `message`: The message to send.
        - `subject`: The subject of the mail.
        - `recipients`: The list of recipients to send the mail to.
        - `cc`: The list of CC recipients.
        - `bcc`: The list of BCC recipients.
        - `in_reply_to`: The Message ID of the mail to reply to.
        - `references`: The list of Message IDs of the mails in the thread.
        - `reply_to`: The list of addresses to recommend recipients to reply to.
        """
        if isinstance(message, Message):
            if not recipient:
                raise ValueError("recipient is required when sending a Message")
            if "subject" in message and subject is None:
                subject = message["subject", 0].data["subject"]
            if "reply" in message:
                reply_id = message["reply", 0].data["id"]
                if subject is None:
                    reply_mail = await self.fetch_mail_of_id(str(reply_id))
                    reply_subject = reply_mail.subject if reply_mail else None
                    subject = f"Re: {reply_subject}" if reply_subject else None
                in_reply_to = str(reply_id) if in_reply_to is None else in_reply_to
                references = [str(reply_id)] if references is None else references
            cc = [cc] if isinstance(cc, str) else cc if cc else []
            bcc = [bcc] if isinstance(bcc, str) else bcc if bcc else []
            for segment in message["cc"]:
                cc.append(segment.data["id"])
            for segment in message["bcc"]:
                bcc.append(segment.data["id"])
            _message = email.mime.multipart.MIMEMultipart()
            _message["From"] = f"{self.bot_info.name} <{self.bot_info.id}>"
            _message["To"] = (
                ", ".join(recipient) if isinstance(recipient, list) else recipient
            )
            if cc:
                _message["Cc"] = ", ".join(cc)
            if bcc:
                _message["Bcc"] = ", ".join(bcc)
            _message["Subject"] = subject or self.bot_info.subject
            if in_reply_to:
                _message["In-Reply-To"] = in_reply_to
            if references:
                _message["References"] = (
                    ", ".join(references)
                    if isinstance(references, list)
                    else references
                )
            elif in_reply_to:
                _message["References"] = in_reply_to
            if reply_to:
                _message["Reply-To"] = (
                    ", ".join(reply_to) if isinstance(reply_to, list) else reply_to
                )
            parts = extract_mail_parts(message)
            for part in parts:
                _message.attach(part)
        else:
            _message = message
        try:
            response = await aiosmtplib.send(
                _message,
                hostname=self.bot_info.smtp.host,
                port=self.bot_info.smtp.port,
                use_tls=self.bot_info.smtp.tls,
                username=self.bot_info.id,
                password=self.bot_info.password.get_secret_value(),
            )
        except Exception as e:
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "failed to connect to SMTP server: "
                    f"{escape_tag(str(e))}"
                ),
            )
            if isinstance(e, aiosmtplib.SMTPRecipientsRefused):
                raise ActionFailed(
                    {
                        i.recipient: aiosmtplib.SMTPResponse(i.code, i.message)
                        for i in e.recipients
                    }
                )
            elif isinstance(e, aiosmtplib.SMTPResponseException):
                raise ActionFailed(
                    {self.bot_info.id: aiosmtplib.SMTPResponse(e.code, e.message)}
                )
            elif isinstance(e, aiosmtplib.SMTPException):
                raise NetworkError()
            else:
                raise e

        log(
            "TRACE",
            (
                f"<y>Bot {escape_tag(self.self_id)}</y> "
                f"mail sent to {escape_tag(str(recipient))}: "
                + escape_tag(str(response))
            ),
        )
        if response[0] != {}:
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "failed to send mail: "
                    f"{escape_tag(str(response[0]))}"
                ),
            )
            raise ActionFailed(response[0])

    async def login(self) -> bool:
        """
        Login to the IMAP server.
        """
        if not self.imap_client:
            raise UninitializedException("IMAP client")
        await self.imap_client.wait_hello_from_server()
        # login to the server
        response = await self.imap_client.login(
            self.bot_info.id, self.bot_info.password.get_secret_value()
        )
        # check if login was successful
        if not response.result == "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    f"error in logging in: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            return False
        # report id to avoid unsafe in 163 mails
        assert self.imap_client.protocol is not None
        response = await asyncio.wait_for(
            self.imap_client.protocol.execute(
                aioimaplib.Command(
                    "ID",
                    self.imap_client.protocol.new_tag(),
                    '("name" "nonebot" "version" "2")',
                )
            ),
            self.imap_client.timeout,
        )
        if not response.result == "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    f"error in reporting client ID: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            return False
        return True

    async def logout(self) -> bool:
        """
        Logout from the IMAP server.
        """
        if not self.imap_client:
            raise UninitializedException("IMAP client")
        response = await self.imap_client.logout()
        if not response.result == "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    f"error in logging out: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            return False
        await self.imap_client.close()
        return True

    async def select_mailbox(
        self, mailbox: str = "INBOX", readonly: bool = False
    ) -> bool:
        """
        Select the mailbox on the IMAP server.

        - `mailbox`: The mailbox to select. Default is "INBOX".
        """
        if not self.imap_client:
            raise UninitializedException("IMAP client")
        if mailbox.startswith('"') and mailbox.endswith('"'):
            mailbox = mailbox[1:-1]
        mailbox = mailbox if " " not in mailbox else f'"{mailbox}"'
        if self.mailbox == mailbox and self.readonly == readonly:
            return True
        response = (
            await self.imap_client.select(mailbox)
            if not readonly
            else await self.imap_client.examine(mailbox)
        )
        if not response.result == "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    f"error in selecting mailbox: "
                    if not readonly
                    else "error in examining mailbox: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            return False
        log(
            "TRACE",
            (
                f"<y>Bot {escape_tag(self.self_id)}</y> "
                f"mailbox {escape_tag(mailbox)} selected ({readonly=}): "
                + escape_bytelines(response.lines)
            ),
        )
        self.mailbox = mailbox
        self.readonly = readonly
        return True

    async def fetch_mail_of_uid(self, mail_uid: str) -> Optional[Mail]:
        """
        Get the mail of the given UID from the current mailbox.

        - `mail_uid`: The UID of the mail to fetch.
        """
        if not self.imap_client:
            raise UninitializedException("IMAP client")
        log(
            "TRACE",
            (
                f"<y>Bot {escape_tag(self.self_id)}</y> "
                f"fetching mail UID: {escape_tag(mail_uid)}"
            ),
        )
        response = await self.imap_client.fetch(mail_uid, "(RFC822)")
        if response.result != "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    f"error in fetching mail UID {mail_uid}: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            raise ActionFailed((self.self_id, response))
        log(
            "TRACE",
            (
                "<y>Bot {escape_tag(self.self_id)}</y> "
                f"mail UID {escape_tag(mail_uid)} fetched"
            ),
        )
        # Parse the mail
        if len(response.lines) < 2:
            log(
                "WARNING",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    f"mail UID {escape_tag(mail_uid)} not found"
                ),
            )
            return None
        mail = parse_byte_mail(response.lines[1])
        return mail

    async def fetch_mail_list(self, key: str = "ALL") -> list[Mail]:
        """
        Fetch mails in current mailbox with the given IMAP search key.

        Note:

        - This will mark the mails as seen.
        - The default key is "ALL".
        - Available keys: https://www.rfc-editor.org/rfc/rfc3501#section-6.4.4
        """
        if not self.imap_client:
            raise UninitializedException("IMAP client")
        response = await self.imap_client.search(key)
        if response.result != "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    "error in fetching seen mails: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            raise ActionFailed((self.self_id, response))
        if len(response.lines) > 0 and response.lines[0]:
            log(
                "TRACE",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    f"seen mail UIDs: {escape_bytelines(response.lines)}"
                ),
            )
        else:
            log(
                "TRACE",
                f"<y>Bot {escape_tag(self.self_id)}</y> no seen mails",
            )
        mail_uids = response.lines[0].decode().split()
        mails: list[Mail] = []
        for mail_uid in mail_uids:
            mail = await self.fetch_mail_of_uid(mail_uid)
            if mail:
                mails.append(mail)
        return mails

    async def fetch_seen_mail_list(self) -> list[Mail]:
        """
        Fetch seen mails in current mailbox.
        """
        return await self.fetch_mail_list("SEEN")

    async def fetch_unseen_mail_list(self) -> list[Mail]:
        """
        Fetch unseen mails in current mailbox and mark them as seen.
        """
        return await self.fetch_mail_list("UNSEEN")

    async def fetch_mail_of_id_in_mailbox(
        self, mail_id: str, mailbox: str = "INBOX"
    ) -> Optional[Mail]:
        """
        Get the mail of the given Message ID from the given mailbox.

        - `mail_id`: The Message ID of the mail to search for.
        - `mailbox`: The mailbox to search in. Default is "INBOX".
        """
        if not self.imap_client:
            raise UninitializedException("IMAP client")
        log(
            "TRACE",
            (
                f"<y>Bot {escape_tag(self.self_id)}</y> "
                f"searching mail ID: {escape_tag(mail_id)} "
                f"in mailbox: {escape_tag(mailbox)}"
            ),
        )
        if not await self.select_mailbox(mailbox):
            return None
        response = await self.imap_client.search(f"HEADER Message-ID {mail_id}")
        log(
            "TRACE",
            (
                f"<y>Bot {escape_tag(self.self_id)}</y> "
                f"mail ID {escape_tag(mail_id)} search result: "
                + escape_bytelines(response.lines)
            ),
        )
        if response.result != "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    f"error in searching mail ID {escape_tag(mail_id)}: "
                    f"{escape_bytelines(response.lines)}"
                    "</bg #f8bbd0></r>"
                ),
            )
            raise ActionFailed((self.self_id, response))
        if not response.lines or not response.lines[0].decode():
            log(
                "TRACE",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    f"mail ID {escape_tag(mail_id)} not found "
                    f"in mailbox {escape_tag(mailbox)}"
                ),
            )
            return None
        mail_uid = response.lines[0].decode()
        return await self.fetch_mail_of_uid(mail_uid)

    async def fetch_mail_of_id(self, mail_id: str) -> Optional[Mail]:
        """
        Get the mail of the given Message ID from the INBOX or Sent mailboxes.

        - `mail_id`: The Message ID of the mail to search for.
        """
        if not self.imap_client or not self.imap_client.protocol:
            raise UninitializedException("IMAP client")
        log(
            "TRACE",
            (
                f"<y>Bot {escape_tag(self.self_id)}</y> "
                f"searching mail ID: {escape_tag(mail_id)}"
            ),
        )
        # try to get mail from INBOX
        mail = await self.fetch_mail_of_id_in_mailbox(mail_id)
        if mail:
            return mail
        # try to get mail from Sent
        response = await self.imap_client.list(
            '""', "*"  # pyright: ignore[reportArgumentType]
        )
        if response.result != "OK":
            log(
                "ERROR",
                (
                    f"<y>Bot {escape_tag(self.self_id)}</y> "
                    "<r><bg #f8bbd0>"
                    "error in listing mailboxes"
                    "</bg #f8bbd0></r>"
                ),
            )
            raise ActionFailed((self.self_id, response))
        sent_mailbox_list = [
            str(i.decode().split(' "/" ')[-1])
            for i in response.lines
            if i.startswith(b"(\\Sent)")
        ]
        for mailbox in sent_mailbox_list:
            mail = await self.fetch_mail_of_id_in_mailbox(mail_id, mailbox)
            if mail:
                break
        # switch back to INBOX
        await self.select_mailbox()
        return mail

    async def handle_event(self, event: Event) -> None:
        if isinstance(event, MessageEvent):
            _check_to_me(self, event)
            await _check_reply(self, event)
        await handle_event(self, event)



from typing import Optional
from typing_extensions import override

from nonebot.utils import escape_tag
from nonebot.compat import model_dump

from nonebot.adapters import Event as BaseEvent

from .model import Mail
from .message import Message


class Event(BaseEvent):

    @override
    def get_event_name(self) -> str:
        return self.get_type()

    @override
    def get_event_description(self) -> str:
        return escape_tag(str(model_dump(self)))

    @override
    def get_message(self) -> Message:
        raise ValueError("Event has no message!")

    @override
    def get_user_id(self) -> str:
        raise ValueError("Event has no context!")

    @override
    def get_session_id(self) -> str:
        raise ValueError("Event has no context!")

    @override
    def is_tome(self) -> bool:
        return False


# Message Event
class MessageEvent(Event, Mail):
    to_me: bool = False

    @override
    def get_type(self) -> str:
        return "message"

    @override
    def is_tome(self) -> bool:
        return self.to_me


class NewMailMessageEvent(MessageEvent):
    reply: Optional[Mail] = None

    @override
    def get_user_id(self) -> str:
        return str(self.sender.id)

    @override
    def get_session_id(self) -> str:
        return str(self.sender.id)

    @override
    def get_event_description(self) -> str:
        return escape_tag(
            f"Message {self.id} from {self.sender.name}<{self.sender.id}>: "
            f"{self.get_message()}"
        )

    @override
    def get_message(self) -> Message:
        return self.message


__all__ = [
    "Event",
    "MessageEvent",
    "NewMailMessageEvent",
]



from typing import Optional
from datetime import datetime

from pydantic import BaseModel

from .message import Message


class User(BaseModel):
    id: str
    name: Optional[str] = None


class Mail(BaseModel):
    id: str
    """
    Mail Message ID
    """
    sender: User
    """
    Mail sender
    """
    subject: str
    """
    Mail subject
    """
    recipients_to: list[User]
    """
    Mail recipients (To)
    """
    recipients_cc: list[User]
    """
    Mail recipients (Cc)
    """
    recipients_bcc: list[User]
    """
    Mail recipients (Bcc)
    """
    date: Optional[datetime]
    """
    Mail date
    """
    timezone: Optional[float]
    """
    Mail timezone offset
    """
    message: "Message"
    """
    Plain text mail message with attachments
    """
    original_message: "Message"
    """
    HTML mail message with attachments
    """
    in_reply_to: Optional[str] = None
    """
    Message ID which this mail is replying to
    """
    references: Optional[list[str]] = None
    """
    Message IDs which this mail references
    """
    reply_to: Optional[list[User]] = None
    """
    Mail addresses to reply to
    """



from typing import Union, Literal, Optional, TypedDict

from aiosmtplib.response import SMTPResponse
from nonebot.exception import AdapterException
from aioimaplib import Response as ImapResponse
from nonebot.exception import ActionFailed as BaseActionFailed
from nonebot.exception import NetworkError as BaseNetworkError
from nonebot.exception import NoLogException as BaseNoLogException
from nonebot.exception import ApiNotAvailable as BaseApiNotAvailable


class MailAdapterException(AdapterException):
    def __init__(self):
        super().__init__("mail")


class Response(TypedDict):
    code: str
    message: str


class ActionFailed(BaseActionFailed, MailAdapterException):
    def __init__(
        self,
        response: Union[tuple[str, ImapResponse], dict[str, SMTPResponse]],
    ):
        self.type: Literal["IMAP", "SMTP"] = "IMAP"
        self.response: dict[str, Response] = {}
        if (
            isinstance(response, tuple)
            and isinstance(response[0], str)
            and isinstance(response[1], ImapResponse)
        ):
            self.type = "IMAP"
            self.response[response[0]] = {
                "code": response[1].result,
                "message": ", ".join(
                    [str(line.decode("utf-8")) for line in response[1].lines]
                ),
            }
        elif isinstance(response, dict) and all(
            isinstance(k, str) and isinstance(v, SMTPResponse)
            for k, v in response.items()
        ):
            self.type = "SMTP"
            for k, v in response.items():
                self.response[k] = {"code": str(v.code), "message": str(v.message)}
        else:
            raise ValueError("Invalid response type")

    @property
    def message(self) -> Optional[str]:
        return None if self.response == {} else str(self.response)

    def __repr__(self) -> str:
        return f"<ActionFailed: {self.type} {self.message}>"

    def __str__(self):
        return self.__repr__()


class NoLogException(BaseNoLogException, MailAdapterException):
    pass


class NetworkError(BaseNetworkError, MailAdapterException):
    pass


class ApiNotAvailable(BaseApiNotAvailable, MailAdapterException):
    pass


class UninitializedException(MailAdapterException):
    def __init__(self, name: str):
        self.name = name

    def __repr__(self) -> str:
        return f"<UninitializedException: {self.name} is not initialized>"

    def __str__(self):
        return self.__repr__()



from io import BytesIO
from pathlib import Path
from dataclasses import dataclass
from collections.abc import Iterable
from typing_extensions import override
from typing import TYPE_CHECKING, Union, Optional, TypedDict

from nonebot.adapters import Message as BaseMessage
from nonebot.adapters import MessageSegment as BaseMessageSegment


class MessageSegment(BaseMessageSegment["Message"]):
    @classmethod
    @override
    def get_message_class(cls) -> type["Message"]:
        return Message

    @override
    def __add__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: Union[str, "MessageSegment", Iterable["MessageSegment"]]
    ) -> "Message":
        return Message(self) + (
            MessageSegment.text(other) if isinstance(other, str) else other
        )

    @override
    def __radd__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, other: Union[str, "MessageSegment", Iterable["MessageSegment"]]
    ) -> "Message":
        return (
            MessageSegment.text(other) if isinstance(other, str) else Message(other)
        ) + self

    @override
    def __str__(self) -> str:
        if self.is_text():
            return self.data.get("text", "")
        params = ", ".join(
            [f"{k}={v!s}" for k, v in self.data.items() if v is not None]
        )
        return f"[{self.type}{':' if params else ''}{params}]"

    @override
    def is_text(self) -> bool:
        return self.type == "text"

    @staticmethod
    def text(text: str) -> "Text":
        return Text("text", {"text": text})

    @staticmethod
    def html(html: str) -> "Html":
        return Html("html", {"html": html})

    @staticmethod
    def attachment(
        data: Union[bytes, BytesIO, Path],
        name: str,
        content_type: Optional[str] = None,
    ) -> "Attachment":
        if isinstance(data, Path):
            data_bytes = data.read_bytes()
        elif isinstance(data, BytesIO):
            data_bytes = data.getvalue()
        else:
            data_bytes = data
        return Attachment(
            "attachment",
            {
                "data": data_bytes,
                "name": name,
                "content_type": content_type,
            },
        )

    @classmethod
    def subject(cls, subject: str) -> "MessageSegment":
        return cls("subject", {"subject": subject})

    @classmethod
    def reply(cls, id: str) -> "MessageSegment":
        return cls("reply", {"id": id})

    @classmethod
    def cc(cls, id: str) -> "MessageSegment":
        return cls("cc", {"id": id})

    @classmethod
    def bcc(cls, id: str) -> "MessageSegment":
        return cls("bcc", {"id": id})


class _TextData(TypedDict):
    text: str


@dataclass
class Text(MessageSegment):
    if TYPE_CHECKING:
        data: _TextData  # pyright: ignore[reportGeneralTypeIssues]

    @override
    def __str__(self) -> str:
        return self.data["text"]

    @override
    def is_text(self) -> bool:
        return True


class _HtmlData(TypedDict):
    html: str


@dataclass
class Html(MessageSegment):
    if TYPE_CHECKING:
        data: _HtmlData  # pyright: ignore[reportGeneralTypeIssues]

    @override
    def __str__(self) -> str:
        return f"[html:{self.data['html']}]"


class _AttachmentData(TypedDict):
    data: bytes
    name: str
    content_type: Optional[str]


@dataclass
class Attachment(MessageSegment):
    if TYPE_CHECKING:
        data: _AttachmentData  # pyright: ignore[reportGeneralTypeIssues]

    @override
    def __str__(self) -> str:
        return f"[attachment:{self.data['name']}, ...]"


class Message(BaseMessage[MessageSegment]):
    @classmethod
    @override
    def get_segment_class(cls) -> type[MessageSegment]:
        return MessageSegment

    @override
    def __add__(
        self, other: Union[str, MessageSegment, Iterable[MessageSegment]]
    ) -> "Message":
        return super().__add__(
            MessageSegment.text(other) if isinstance(other, str) else other
        )

    @override
    def __radd__(
        self, other: Union[str, MessageSegment, Iterable[MessageSegment]]
    ) -> "Message":
        return super().__radd__(
            MessageSegment.text(other) if isinstance(other, str) else other
        )

    @staticmethod
    @override
    def _construct(msg: str) -> Iterable[MessageSegment]:
        if msg:
            yield Text("text", {"text": msg})

    @override
    def __str__(self) -> str:
        result = ""
        for seg in self:
            result += (
                str(seg)
                if seg.type != "attachment"
                else f"[attachment:{seg.data['name']}]"
            )
        return result

    @override
    def __repr__(self) -> str:
        return f"{self.__str__()!r}"



from pydantic import Field, EmailStr, BaseModel, SecretStr


class HostInfo(BaseModel):
    host: str = Field(alias="host")
    port: int = Field(alias="port")
    tls: bool = Field(alias="tls")


class BotInfo(BaseModel):
    id: EmailStr = Field(alias="id")
    name: str = Field(alias="name")
    password: SecretStr = Field(alias="password")
    subject: str = Field(alias="subject")
    imap: HostInfo = Field(alias="imap")
    smtp: HostInfo = Field(alias="smtp")


class Config(BaseModel):
    mail_bots: list[BotInfo] = Field(default_factory=list)

    class Config:
        extra = "ignore"



import html
import base64
import email.message
import email.encoders
import email.mime.base
import email.mime.text

import mailparser
from nonebot.utils import escape_tag

from .model import Mail, User
from .message import Message, Attachment, MessageSegment


def escape_bytelines(s: list[bytearray]) -> str:
    """
    Escape a list of bytearrays to a string.

    - `s`: The list of bytearrays to escape.
    """
    return f'[{escape_tag(", ".join([i.decode() for i in s]))}]'


def extract_mail_parts(message: Message) -> list[email.message.EmailMessage]:
    """
    Extract email parts from a Message object.

    - `message`: The Message object to extract.
    """
    text: str = ""
    attachments = []
    contains_html = any(segment.type == "html" for segment in message)
    for segment in message:
        if segment.type == "text":
            text += (
                segment.data["text"]
                if not contains_html
                else html.escape(segment.data["text"])
            )
        elif segment.type == "html":
            text += segment.data["html"]
        elif segment.type == "attachment":
            attachments.append(segment)
    parts = []
    if contains_html:
        parts.append(email.mime.text.MIMEText(text, "html"))
    else:
        parts.append(email.mime.text.MIMEText(text))
    for attachment in attachments:
        if attachment.data["content_type"] and "/" in attachment.data["content_type"]:
            main_type, sub_type = attachment.data["content_type"].split("/")
        else:
            main_type, sub_type = "application", "octet-stream"
        part = email.mime.base.MIMEBase(main_type, sub_type)
        part.set_payload(attachment.data["data"])
        email.encoders.encode_base64(part)
        part.add_header(
            "Content-Disposition",
            f'attachment; filename="{attachment.data["name"]}"',
        )
        parts.append(part)
    return parts


def parse_attachment(attachment) -> Attachment:
    """
    Parse the attachment and return the Attachment object.

    - `attachment`: The attachment to parse.
    """

    data = (
        base64.b64decode(attachment["payload"])
        if attachment["binary"]
        else attachment["payload"]
    )

    if isinstance(data, str):
        data = data.encode()

    return MessageSegment.attachment(
        data,
        attachment["filename"],
        attachment["mail_content_type"],
    )


def parse_byte_mail(byte_mail: bytes) -> Mail:
    """
    Parse the mail and return the Mail object.

    - `byte_mail`: The byte mail to parse.
    """

    def parse_user(user):
        return User(
            id=user[1],
            name=user[0],
        )

    mail = mailparser.parse_from_bytes(byte_mail)

    sender = User(
        id=mail.from_[0][1],
        name=mail.from_[0][0],
    )
    if isinstance(mail.subject, str):
        subject = mail.subject
    else:
        subject = str(mail.subject[0]) if mail.subject else ""
    recipients_to = [parse_user(u) for u in mail.to]
    recipients_cc = [parse_user(u) for u in mail.headers.get("Cc", [])]
    recipients_bcc = [parse_user(u) for u in mail.headers.get("Bcc", [])]
    reply_to = [parse_user(u) for u in mail.headers.get("Reply-To", [])]
    message = Message(
        [MessageSegment.text(text) for text in mail.text_plain]
    ) + Message([parse_attachment(attachment) for attachment in mail.attachments])
    original_message = Message(
        [MessageSegment.html(html) for html in mail.text_html]
    ) + Message(
        [parse_attachment(attachment) for attachment in mail.attachments],
    )

    if str(message) == "":
        # no plain text provided by the mail
        message = Message([MessageSegment.text("")])

    return Mail(
        id=str(mail.message_id),
        sender=sender,
        subject=subject,
        recipients_to=recipients_to,
        recipients_cc=recipients_cc,
        recipients_bcc=recipients_bcc,
        date=mail.date,
        timezone=float(mail.timezone) if mail.timezone else None,
        message=message,
        original_message=original_message,
        in_reply_to=str(mail.in_reply_to) if mail.in_reply_to else None,
        references=(
            [str(reference) for reference in mail.references]
            if mail.references
            else None
        ),
        reply_to=reply_to if reply_to else None,
    )
